-- First action slot used by this addon should be 109
local ACTION_OFFSET = 108

-- States used to control the buttons
local HIDE_STATE   = 0 -- Hide the buttons
local SHOW_STATE   = 1 -- Show the buttons when a frame is clicked
local CONFIG_STATE = 2 -- Show the buttons via slash command for configuration

-- Amount of time in seconds to hide the buttons after mousing away from them
local HIDE_DELAY   = 2

-- Create slash commands to show/hide the buttons for configuration
SLASH_UNITACTIONPOPUP1 = "/unitactionpopup"
SLASH_UNITACTIONPOPUP2 = "/uap"
SlashCmdList["UNITACTIONPOPUP"] = function(text)
  if InCombatLockdown() then
    DEFAULT_CHAT_FRAME:AddMessage("You cannot configure UnitActionPopup in combat.")
    return
  end
  
  if text == "config" then
    UnitActionPopupHeader:SetAttribute("state", CONFIG_STATE)
    -- There is a bug in the UI rendering system where a child of a hidden frame
    -- will not be shown when it is re-parented to a visible frame. We must
    -- manually show the frame after setting it's state. Remove this if the bug
    -- is fixed.
    UnitActionPopupHeader:Show()
  elseif text == "hide" then
    UnitActionPopupHeader:SetAttribute("state", HIDE_STATE)
  else
    -- Using two AddMessages to make for better scrolling
    DEFAULT_CHAT_FRAME:AddMessage([[
Usage:
  /unitactionpopup |cff33ff33command|r
  /uap |cff33ff33command|r]]
    )
    DEFAULT_CHAT_FRAME:AddMessage([[
Commands:
  |cff33ff33config|r - Show the buttons for configuration
  |cff33ff33hide|r - Hide the buttons]]
    )
  end
end

local function SetManyAttributes(frame, ...)
  for i = 1, select("#", ...), 2 do
    frame:SetAttribute(select(i, ...))
  end
end

UnitActionPopup = {}

function UnitActionPopup:ButtonOnLoad(button)
  -- The ID was only used to make the buttons easily-identifiable by this
  -- function. Retrieve it and then clear it so the button will behave as
  -- expected.
  local id = button:GetID()
  button:SetID(0)
  
  -- Set up various attributes for the button
  SetManyAttributes(button,
    "action",        id + ACTION_OFFSET,
    "hidestates",    HIDE_STATE,
    "checkselfcast", nil
  )
  
  -- Add the button to the header
  button:GetParent():SetAttribute("addchild", button)
  
  -- Call the built-in OnLoad function
  ActionButton_OnLoad()
end

local function FixedSetClampRectInsets(frame, left, right, top, bottom)
  frame:SetClampRectInsets(left, -right, -top, bottom)
end

function UnitActionPopup:OnLoad(frame)
  -- No ClampRectInsets in the XML schema yet, and the widget method is bugged
  FixedSetClampRectInsets(frame, -36, -36, -36, -36)
  
  frame:RegisterEvent("PLAYER_REGEN_DISABLED")
  
  SetManyAttributes(frame,
    "useparent-unit",       true,
    
    -- Anchor the frame to the cursor in state 1, the center of the screen in
		-- state 2, and off the left side of the screen in state 0.
    "headofsx",            format("%d:-18;0", HIDE_STATE),
    "headofsrelpoint",     format("%d:cursor;%d:LEFT", SHOW_STATE, HIDE_STATE),
    "headparent",          format("%d,%d:uiparent", CONFIG_STATE, HIDE_STATE),
    "headparent-uiparent", UIParent,
    
    -- Go into the show state when the anchor is clicked
    "statemap-anchor",      SHOW_STATE,
    -- Hide the frame again after a delay if it was shown by a click
    "delaystatemap-anchor", HIDE_STATE,
    "delaytimemap-anchor",  HIDE_DELAY,
    "delayhovermap-anchor", "true"
  )
  
  -- Set up the anchor frames
  self:AddAnchors(frame)
end

function UnitActionPopup:OnEvent(frame, event, ...)
  if event == "PLAYER_REGEN_DISABLED" then
    -- Hide buttons when entering combat while in config mode
    if frame:GetAttribute("state") == CONFIG_STATE then
      frame:SetAttribute("state", HIDE_STATE)
    end
  end
end

-- This is a list of all the frames for which we will create anchors
local unitFrames = {
  PlayerFrame,
  PetFrame,
  TargetFrame,
  TargetofTargetFrame,
  PartyMemberFrame1,
  PartyMemberFrame2,
  PartyMemberFrame3,
  PartyMemberFrame4,
  PartyMemberFrame1PetFrame,
  PartyMemberFrame2PetFrame,
  PartyMemberFrame3PetFrame,
  PartyMemberFrame4PetFrame,
}
function UnitActionPopup:AddAnchors(header)
  for _, unitFrame in pairs(unitFrames) do
    self:AddAnchor(unitFrame, header)
  end
end

function UnitActionPopup:AddAnchor(parent, header)
  -- Create the anchor
  local anchor = CreateFrame(
    "Button",
    parent:GetName().."UAPAnchor",
    parent,
    "SecureAnchorButtonTemplate"
  )
  -- Make sure it'll be above most other frames
  anchor:SetFrameStrata("DIALOG")
  
  -- Change the click behavior of the unit frame
  SetManyAttributes(parent,
    -- Notice the lack of *, making it so that holding a modifier key when
    -- clicking will do the normal action of the unit frame
    "type1",       "click",
    "clickbutton",  anchor
  )
  
  SetManyAttributes(anchor,
    "useparent-unit", true,
    "anchorchild",    UnitActionPopupHeader,
    "childreparent",  "true",
    "childstate",     "^"..SHOW_STATE
  )
end
